import numpy as np

from bisect import bisect_left
import cvxopt

from intvalpy.RealInterval import Interval

from intvalpy.utils import intersection, infinity
from intvalpy.linear.square_system import Gauss
from intvalpy.linear.system_properties import Tol


cvxopt.solvers.options['show_progress'] = False
cvxopt.solvers.options['glpk'] = {'msg_lev': 'GLP_MSG_OFF'}


def _Rohn_Tol(A, b, nu=None):

    def solve_lp(c, A_ub, b_ub):
        asmatrix = lambda A: A if isinstance(A, cvxopt.matrix) else cvxopt.matrix(A)
        args = [asmatrix(c), asmatrix(A_ub), asmatrix(b_ub)]

        result = cvxopt.solvers.lp(*args, solver='glpk')
        if 'optimal' not in result['status']:
            raise Exception("Неизвестная ошибка.")
        return np.array(result['x']).reshape(c.shape[0])

    def algo(A, b, nu):
        A_ub = np.zeros((2*(n+m), 2*m), dtype=np.float64)
        b_ub = np.zeros(2*(n+m), dtype=np.float64)

        A_ub[:n, :m] = A.b
        A_ub[:n, m:] = -A.a

        A_ub[n:2*n, :m] = -A.a
        A_ub[n:2*n, m:] = A.b
        A_ub[2*n:] = -np.eye(2*m)

        b_ub[:n] = b.b
        b_ub[n:2*n] = -b.a

        c = np.zeros(2*m, dtype=np.float64)
        c[nu], c[nu+m] = 1, -1

        inf = solve_lp(c, A_ub, b_ub) @ c
        sup = -(solve_lp(-c, A_ub, b_ub) @ -c)

        return inf, sup

    n, m = A.shape
    tol = Tol(A, b, maxQ=True)
    if tol[2] < 0:
        raise Exception('Допусковое множество решений пусто!')

    _inf, _sup = [], []
    if nu is None:
        for k in range(m):
            inf, sup = algo(A, b, k)
            _inf.append(inf)
            _sup.append(sup)
    else:
        inf, sup = algo(A, b, nu)
        _inf.append(inf)
        _sup.append(sup)

    return Interval(_inf, _sup, sortQ=False)


def _Rohn_Uni(A, b, tol=1e-12, maxiter=2000):

    Amid = np.array(A.mid, dtype=np.float64)
    Ac_plus = np.linalg.inv(Amid.T @ Amid) @ Amid.T
    A = Ac_plus @ A.copy
    b = Ac_plus @ b.copy

    n, m = A.shape
    assert m <= n, 'Количество строк должно быть не меньше, чем количество столбцов!'

    Amid = np.array(A.mid, dtype=np.float64)
    R = np.linalg.inv(Amid.T @ Amid) @ Amid.T
    x0 = R @ b.mid

    G = abs(np.eye(m) - R @ Amid) + abs(R) @ A.rad
    g = abs(R @ (Amid @ x0 - b.mid)) + abs(R) @ (A.rad @ abs(x0) + b.rad)

    result = np.zeros(m)
    error = infinity
    nit = 0
    while error >= tol and nit <= maxiter:
        d = np.copy(result)
        result = G @ d + g + tol
        error = np.amax(abs(result-d))
        nit += 1
    return Interval(x0-result, x0+result)


def Rohn(A, b, tol=1e-12, maxiter=2000, consistency='uni'):
    """
    Метод Дж. Рона для переопределённых ИСЛАУ.

    Parameters:
                A: Interval
                    Матрица ИСЛАУ.

                b: Interval
                    Вектор правой части ИСЛАУ.

    Optional Parameters:
                tol: float
                    Погрешность для остановки итерационного процесса.

                maxiter: int
                    Максимальное количество итераций.

                consistency: str
                    Параметр указывает какое множество решений (объединённое или
                    допусковое) будет выведено в ответе.

    Returns:
                out: Interval
                    Возвращается интервальный вектор решений.
    """
    if consistency == 'uni':
        return _Rohn_Uni(A, b, tol=tol, maxiter=maxiter)
    elif consistency == 'tol':
        return _Rohn_Tol(A, b, nu=None)
    else:
        raise Exception('Нахождение данного множества решений не предусмотрено.')


def PSS(A, b, tol=1e-12, maxiter=2000, nu=None):
    """
    A hybrid method of crushing PSS solutions designed to find an external optimal estimate
    of the united set of solutions of interval systems of linear algebraic equations (ISLAE) A x = b.
    Since the task is NP-hard, the process can be stopped by the number of iterations completed.
    PSS methods are consistently guaranteeing, i.e. when the process is interrupted at any number of iterations,
    an approximate estimate of the solution satisfies the required estimation method.

    Returns the formal solution of the interval system of linear equations.
    If it is not necessary to evaluate all the components, then any nu component can be evaluated.


    Parameters:

        A: Interval
            The input interval matrix of ISLAE, which can be either square or rectangular.

        b: Interval
            The interval vector of the right part of the ISLAE.

        tol: float, optional
            The error that determines when further crushing of the bars is unnecessary,
            i.e. their width is "close enough" to zero, which can be considered exactly zero.

        maxiter: int, optional
            The maximum number of iterations.

        nu: int, optional
            Choosing the number of the component along which the set of solutions is evaluated.

    Returns:

        out: Interval
            Returns an interval vector, which, after substituting into the system of equations
            and performing all operations according to the rules of arithmetic and analysis,
            turns the equations into true equalities.
    """


    class KeyWrapper:
        def __init__(self, iterable, key):
            self.it = iterable
            self.key = key

        def __getitem__(self, i):
            return self.key(self.it[i])

        def __len__(self):
            return len(self.it)


    def kdiv(a, b, num_func):
        """
        Деление в арифметике Кахана.
        """
        func0 = lambda a, b: Interval(-infinity, a.b/b.b, sortQ=False)
        func1 = lambda a, b: Interval(a.b/b.a, infinity, sortQ=False)
        func2 = lambda a, b: Interval([-infinity, a.b/b.a],
                                      [a.b/b.b, infinity], sortQ=False)

        func3 = lambda a, b: Interval(a.a/b.b, infinity, sortQ=False)
        func4 = lambda a, b: Interval(-infinity, a.a/b.a, sortQ=False)
        func5 = lambda a, b: Interval([-infinity, a.a/b.b],
                                      [a.a/b.a, infinity], sortQ=False)

        kdiv_result = [func0, func1, func2, func3, func4, func5]

        if 0 in a:
            return Interval(-infinity, infinity, sortQ=False)
        elif a.b<0:
            return kdiv_result[num_func](a, b)
        elif 0<a.a:
            return kdiv_result[3+num_func](a, b)
        else:
            raise Exception('Деление не определено!')


    def Omega(bAr, _nu, Anu, index_classic_div, index_kahan_div, num_func):
        try:
            div = bAr[index_classic_div]/Anu[index_classic_div]
        except:
            div = []
        pdiv = []

        for k in index_kahan_div:
            pdiv.append(kdiv(bAr[k], Anu[k], num_func[k]))

        try:
            result = intersection(div[0], V[_nu])
            for el in div[1:]:
                result = intersection(result, el)
        except:
            result = intersection(pdiv[0], V[_nu])

        for el in pdiv:
            result = intersection(result, el)

        if 2 in num_func:
            # if np.isnan(result.b).all():
            if np.isnan(np.array(result.b, dtype=np.float64)).all():
                return infinity
            elif np.isnan(np.array(result.b, dtype=np.float64)).any():
                return np.max(result.a)
            else:
                return np.min(result.a)
        else:
            if np.isnan(np.array(result.b, dtype=np.float64)):
                return infinity
            else:
                return result.a


    n, m = A.shape
    midA = A.mid
    radA = A.rad

    def StartBar(A, b):
        if n > m:
            try:
                V = Rohn(A, b)
            except:
                V = Interval([-10**14]*m, [10**14]*m, sortQ=False)
        else:
            try:
                V = Gauss(A, b)
            except:
                V = Interval([-10**14]*m, [10**14]*m, sortQ=False)

        return V

    A_ub = np.zeros((2*(n+m)+m, m))
    A_ub[2*(n+m) : ] = -np.eye(m)

    b_ub = np.zeros(2*(n+m)+m)
    p = np.zeros(2*m)

    def solve_lp(c, A_ub, b_ub):
        asmatrix = lambda A: A if isinstance(A, cvxopt.matrix) else cvxopt.matrix(A)
        args = [asmatrix(c), asmatrix(A_ub), asmatrix(b_ub)]

        result = cvxopt.solvers.lp(*args, solver='glpk')
        if 'optimal' not in result['status']:
            raise Exception("LP optimum not found: %s" % result['status'])
        return np.array(result['x']).reshape(c.shape[0])

    def ExactOmega(midA, radA, b, Q):
        S = np.eye(m) * np.insert(np.sign(Q.a), _nu, np.sign(V[_nu].a))
        midAS = midA @ S

        p[:m] = np.insert(-Q.a, _nu, 1e15)
        p[m:] = np.insert(Q.b, _nu, 1e15)

        c = np.zeros(m, dtype='float64')
        c[_nu] = S[_nu, _nu]

        A_ub[:n] = midAS - radA
        A_ub[n:2*n] = -midAS - radA
        A_ub[2*n:2*n+m] = -S
        A_ub[2*n+m:2*(n+m)] = S

        b_ub[:n] = b.b
        b_ub[n:2*n] = -b.a
        b_ub[2*n:2*(n+m)] = p

        try:
            result = solve_lp(c, A_ub, b_ub)
            return result @ c
        except Exception:
            pass

        S[_nu, _nu] = np.sign(V[_nu].b)
        midAS = midA @ S

        c[_nu] = S[_nu, _nu]
        A_ub[:n] = midAS - radA
        A_ub[n:2*n] = -midAS - radA
        A_ub[2*n:2*n+m] = -S
        A_ub[2*n+m:2*(n+m)] = S

        try:
            result = solve_lp(c, A_ub, b_ub)
            return result @ c
        except Exception:
            return 10**15

    def algo(_nu):
        WorkListA = A.copy; del(WorkListA[:, _nu])
        WorkListb = b.copy

        Q = V.copy
        q, omega = Q[_nu].a, Q[_nu].b
        del(Q[_nu])

        Anu = A[:, _nu]
        index_classic_div = np.where(Anu.a*Anu.b > 0)[0]
        index_kahan_div = np.array([k for k in range(len(Anu.a)) if not (k in index_classic_div)])

        num_func = np.zeros(len(Anu), dtype='int32')
        for index in index_kahan_div:
            if not Anu[index].a:
                 num_func[index] = 0
            elif not Anu[index].b:
                 num_func[index] = 1
            else:
                 num_func[index] = 2

        L = [(Q, q)]
        Ar = WorkListA @ Q

        item = 0
        nit = 0
        while np.amax(Q.wid) >= tol and nit <= maxiter:
            k = np.argmax(Q.wid)
            Q1 = L[item][0].copy
            Q2 = L[item][0].copy

            if Q[k].a:
                if -2 < Q[k].b / Q[k].a < -1/2:
                    Q1[k], Q2[k] = Interval(Q[k].a, 0, sortQ=False), Interval(0, Q[k].b, sortQ=False)
                else:
                    Q1[k], Q2[k] = Interval(Q[k].a, Q[k].mid, sortQ=False), Interval(Q[k].mid, Q[k].b, sortQ=False)
            else:
                Q1[k], Q2[k] = Interval(Q[k].a, Q[k].mid, sortQ=False), Interval(Q[k].mid, Q[k].b, sortQ=False)

            del L[item]

            matmul = True
            if 0 in Q1:
                QA = Q[k] * WorkListA[:, k]
                Ar = Ar + QA.opp
                matmul = False

                Ar1 = Ar + Q1[k] * WorkListA[:, k]
                bAr = WorkListb - Ar1
                q1 = Omega(bAr, _nu, Anu, index_classic_div, index_kahan_div, num_func)

                if q1 < omega:
                    newcol = (Q1, q1, Ar1)
                    bslindex = bisect_left(KeyWrapper(L, key=lambda c: c[1]), newcol[1])
                    L.insert(bslindex, newcol)
                    eta1 = infinity
                else:
                    eta1 = infinity
            else:
                q1 = ExactOmega(midA, radA, b, Q1)

                if q1 < omega:
                    newcol = (Q1, q1)
                    bslindex = bisect_left(KeyWrapper(L, key=lambda c: c[1]), newcol[1])
                    L.insert(bslindex, newcol)
                    eta1 = q1
                else:
                    eta1 = infinity

            if 0 in Q2:
                if matmul:
                    QA = Q[k] * WorkListA[:, k]
                    Ar = Ar + QA.opp

                Ar2 = Ar + Q2[k] * WorkListA[:, k]
                bAr = WorkListb - Ar2
                q2 = Omega(bAr, _nu, Anu, index_classic_div, index_kahan_div, num_func)

                if q2 < omega:
                    newcol = (Q2, q2, Ar2)
                    bslindex = bisect_left(KeyWrapper(L, key=lambda c: c[1]), newcol[1])
                    L.insert(bslindex, newcol)
                    eta2 = infinity
                else:
                    eta2 = infinity
            else:
                q2 = ExactOmega(midA, radA, b, Q2)

                if q2 < omega:
                    newcol = (Q2, q2)
                    bslindex = bisect_left(KeyWrapper(L, key=lambda c: c[1]), newcol[1])
                    L.insert(bslindex, newcol)
                    eta2 = q2
                else:
                    eta2 = infinity

            eta = eta1 if eta1 <= eta2 else eta2
            if omega > eta:
                omega = eta
                L = [l for l in L if l[1] <= omega]

            item = 0
            while True:
                try:
                    Q, q, Ar = L[item]
                    break
                except IndexError:
                    return L[0][1]
                except:
                    item += 1

            nit += 1

        return L[0][1]

    inf = []
    sup = []
    if nu is None:
        for endint in [1, -1]:
            b = endint * b.copy
            V = StartBar(A, b)
            for _nu in range(A.shape[1]):
                if endint == -1:
                    sup.append(endint * algo(_nu))
                else:
                    inf.append(algo(_nu))
    else:
        _nu = nu
        for endint in [1, -1]:
            b = endint * b.copy
            V = StartBar(A, b)
            if endint == -1:
                sup.append(endint * algo(_nu))
            else:
                inf.append(algo(_nu))

    return Interval(inf, sup, sortQ=False)



import numpy as np

import mpmath
import math
from mpmath import mp, mpf

import itertools


def wrap_to_interval(func):
    def interval(x, y):
        if isinstance(y, ARITHMETICS):
            return func(x, y)
        else:
            return func(x, SingleInterval(y, y))
    return interval


class BaseTools(object):

    def __init__(self, left, right):
        self._a, self._b = left, right

    @property
    def a(self):
        return self._a

    @property
    def b(self):
        return self._b

    @property
    def inf(self):
        return self._a

    @property
    def sup(self):
        return self._b

    @property
    def copy(self):
        return type(self)(self._a, self._b)

    def to_float(self):
        return type(self)(np.float64(self._a), np.float64(self._b))

    @property
    def wid(self):
        """Возвращает ширину интервала."""
        return abs(self._b - self._a)

    @property
    def rad(self):
        """Возвращает радиус интервала."""
        return 1/2 * self.wid

    @property
    def mid(self):
        """Возвращает середину интервала."""
        return 1/2 * (self._b + self._a)

    @property
    def mig(self):
        """Возвращает мигнитуду интервала."""
        if 0 in self:
            return 0.0
        else:
            return min(abs(self._a), abs(self._b))

    @property
    def mag(self):
        """Возвращает магнитуду интервала."""
        return max(abs(self._a), abs(self._b))

    @property
    def dual(self):
        if self._b <= self._a:
            return ClassicalArithmetic(self._b, self._a)
        else:
            return KaucherArithmetic(self._b, self._a)

    @property
    def pro(self):
        if isinstance(self, ClassicalArithmetic):
            return ClassicalArithmetic(self._a, self._b)
        else:
            return ClassicalArithmetic(min(self._a, self._b), max(self._a, self._b))

    @property
    def opp(self):
        if self._b <= self._a:
            return ClassicalArithmetic(-self._a, -self._b)
        else:
            return KaucherArithmetic(-self._a, -self._b)

    @property
    def inv(self):
        if self._b <= self._a:
            return ClassicalArithmetic(1/self._a, 1/self._b)
        else:
            return KaucherArithmetic(1/self._a, 1/self._b)

    def __repr__(self):
        return "'[%.6g, %.6g]'" % (self._a, self._b)

    # Унарные числовые операции
    def __neg__(self):
        return type(self)(-self._b, -self._a)

    def __abs__(self):
        return type(self)(self.mig, self.mag)

    @wrap_to_interval
    def __contains__(self, other):
        return (self.a <= other.a) and (other.b <= self.b)

    @wrap_to_interval
    def __it__(self, other):
        return (self.a < other.a) and (self.b < other.b)

    @wrap_to_interval
    def __le__(self, other):
        return (self.a <= other.a) and (self.b <= other.b)

    @wrap_to_interval
    def __eq__(self, other):
        return (self.a == other.a) and (self.b == other.b)

    @wrap_to_interval
    def __ne__(self, other):
        return (self.a != other.a) or (self.b != other.b)

    @wrap_to_interval
    def __gt__(self, other):
        return (self.a > other.a) and (self.b > other.b)

    @wrap_to_interval
    def __ge__(self, other):
        return (self.a >= other.a) and (self.b >= other.b)

    # def __pow__(self, other):
    #     return np.exp(other * np.log(self))

    def __array_ufunc__(self, *args):
        cls = type(self)
        if args[0].__name__ in ['add']:
            return self.__radd__(args[2])

        elif args[0].__name__ in ['subtract']:
            return cls(-self._b, -self._a).__radd__(args[2])

        elif args[0].__name__ in ['multiply']:
            return self.__rmul__(args[2])

        elif args[0].__name__ in ['true_divide']:
            return cls(1/self._b, 1/self._a).__rmul__(args[2])

        elif args[0].__name__ in ['sqrt']:
            return np.exp(1/2 * np.log(self))

        elif args[0].__name__ in ['exp']:
            try:
                return cls(math.exp(self._a), math.exp(self._b))
            except OverflowError:
                return cls(float('inf'), float('inf'))

        elif args[0].__name__ in ['log']:
            pro = self.pro
            _max, _min = max(pro.a, 0), pro.b
            if _max <= _min:
                inf = float('-inf') if _max==0 else math.log(_max)
                sup = math.log(_min)
                return cls(inf, sup) if self._a <= self._b else cls(sup, inf)
            else:
                return cls(float('nan'), float('nan'))

        elif args[0].__name__ in ['sin']:
            x = self.pro
            sin_inf, sin_sup = math.sin(x.a), math.sin(x.b)

            if math.ceil((x.a + math.pi / 2) / (2 * math.pi)) <= math.floor((x.b + math.pi / 2) / (2 * math.pi)):
                inf = -1.0
            else:
                inf = min(sin_inf, sin_sup)

            if math.ceil((x.a - math.pi / 2) / (2 * math.pi)) <= math.floor((x.b - math.pi / 2) / (2 * math.pi)):
                sup = 1.0
            else:
                sup = max(sin_inf, sin_sup)

            if self._a <= self._b:
                return cls(inf, sup)
            else:
                return cls(sup, inf)

        elif args[0].__name__ in ['cos']:
            x = self.pro
            cos_inf, cos_sup = math.cos(x.a), math.cos(x.b)

            if math.ceil((x.a - math.pi) / (2 * math.pi)) <= math.floor((x.b - math.pi) / (2 * math.pi)):
                inf = -1.0
            else:
                inf = min(cos_inf, cos_sup)

            if math.ceil(x.a / (2 * math.pi)) <= math.floor(x.b / (2 * math.pi)):
                sup = 1.0
            else:
                sup = max(cos_inf, cos_sup)

            if self._a <= self._b:
                return cls(inf, sup)
            else:
                return cls(sup, inf)

        else:
            raise Exception("Расчёт функции {} не предусмотрен!".format(args[0].__name__))


class ClassicalArithmetic(BaseTools):

    def __add__(self, other):
        if isinstance(other, ARITHMETICS):
            return type(other)(self._a + other._a, self._b + other._b)
        else:
            return other + self

    def __sub__(self, other):
        if isinstance(other, ARITHMETICS):
            return type(other)(self._a - other._b, self._b - other._a)
        else:
            return -other + self

    def __mul__(self, other):
        if isinstance(other, ClassicalArithmetic):
            mul = np.array([self._a*other._a,
                            self._a*other._b,
                            self._b*other._a,
                            self._b*other._b])
            return ClassicalArithmetic(np.min(mul), np.max(mul))

        elif isinstance(other, KaucherArithmetic):
            _selfInfPlus, _selfSupPlus = max(self._a, 0), max(self._b, 0)
            _selfInfMinus, _selfSupMinus = _selfInfPlus - self._a, _selfSupPlus - self._b

            _otherInfPlus, _otherSupPlus = max(other.a, 0), max(other.b, 0)
            _otherInfMinus, _otherSupMinus = _otherInfPlus - other.a, _otherSupPlus - other.b

            return KaucherArithmetic(_selfInfPlus*_otherInfPlus + _selfSupMinus*_otherSupMinus - _selfSupPlus*_otherInfMinus - _selfInfMinus*_otherSupPlus, \
                                    _selfSupPlus*_otherSupPlus + _selfInfMinus*_otherInfMinus - _selfInfPlus*_otherSupMinus - _selfSupMinus*_otherInfPlus)

        else:
            return other * self

    def __truediv__(self, other):

        if isinstance(other, ClassicalArithmetic):
            assert not (0 in other), 'Нельзя делить на нуль содержащие интервалы!'

            div = np.array([self._a/other._a,
                            self._a/other._b,
                            self._b/other._a,
                            self._b/other._b])
            return ClassicalArithmetic(div.min(), div.max())

        if isinstance(other, KaucherArithmetic):
            assert not (0 in other), 'Нельзя делить на нуль содержащие интервалы!'
            return self.__mul__(KaucherArithmetic(1/other.b, 1/other.a))

        else:
            return 1/other * self

    def __pow__(self, other):
        if isinstance(other, (int, np.int32, np.int64)) and other >= 0:
            inf, sup = self._a**other, self._b**other
            if other % 2 == 0 and 0 in self:
                return ClassicalArithmetic(0, max(inf, sup))
            else:
                return ClassicalArithmetic(min(inf, sup), max(inf, sup))
        elif self >= 0:
            return np.exp(other * np.log(self))
        else:
            raise ValueError('Если основание содержит отрицательные числа, то степень может быть только натуральным числом.')

    def __radd__(self, other):
        if isinstance(other, single_type):
            return ClassicalArithmetic(self._a + other, self._b + other)
        else:
            return ArrayInterval(np.vectorize(lambda o: o + self)(other))

    def __rsub__(self, other):
        if isinstance(other, single_type):
            return ClassicalArithmetic(other - self._b, other - self._a)
        else:
            return ArrayInterval(np.vectorize(lambda o: o - self)(other))

    def __rmul__(self, other):
        def fmul(x, y):
            mul = np.array([x*y.a, x*y.b])
            return ClassicalArithmetic(min(mul), max(mul))

        if isinstance(other, single_type):
            return fmul(other, self)
        else:
            return ArrayInterval(np.vectorize(lambda o: o * self)(other))

    def __rtruediv__(self, other):
        return other * ClassicalArithmetic(1/self._b, 1/self._a)

    #  Арифметические операторы присваивания
    def __iadd__(self, other):
        other = Interval(other)
        if isinstance(other, ClassicalArithmetic):
            self._a, self._b = self._a + other._a, self._b + other._b
            return self
        else:
            raise SyntaxError("It is not possible to change the type of the variable.")

    def __isub__(self, other):
        other = Interval(other)
        if isinstance(other, ClassicalArithmetic):
            self._a, self._b = self._a - other._b,  self._b - other._a
            return self
        else:
            raise SyntaxError("It is not possible to change the type of the variable.")

    def __imul__(self, other):
        other = Interval(other)
        if isinstance(other, ClassicalArithmetic):
            mul = np.array([self._a*other._a,
                            self._a*other._b,
                            self._b*other._a,
                            self._b*other._b])
            self._a, self._b = mul.min(), mul.max()
            return self
        else:
            raise SyntaxError("It is not possible to change the type of the variable.")

    def __itruediv__(self, other):
        other = Interval(other)
        if isinstance(other, ClassicalArithmetic):
            assert not (0 in other), 'Нельзя делить на нуль содержащие интервалы!'

            div = np.array([self._a/other._a,
                            self._a/other._b,
                            self._b/other._a,
                            self._b/other._b])
            self._a, self._b = div.min(), div.max()
            return self
        else:
            raise SyntaxError("It is not possible to change the type of the variable.")


class KaucherArithmetic(BaseTools):

    #     Арифметические операторы
    def __add__(self, other):
        if isinstance(other, ARITHMETICS):
            return KaucherArithmetic(self._a + other.a, self._b + other.b)
        else:
            return other + self

    def __sub__(self, other):
        if isinstance(other, ARITHMETICS):
            return KaucherArithmetic(self._a - other.b, self._b - other.a)
        else:
            return -other + self

    def __mul__(self, other):
        if isinstance(other, ARITHMETICS):
            _selfInfPlus, _selfSupPlus = max(self._a, 0), max(self._b, 0)
            _selfInfMinus, _selfSupMinus = _selfInfPlus - self._a, _selfSupPlus - self._b

            _otherInfPlus, _otherSupPlus = max(other.a, 0), max(other.b, 0)
            _otherInfMinus, _otherSupMinus = _otherInfPlus - other.a, _otherSupPlus - other.b

            return KaucherArithmetic(max(_selfInfPlus * _otherInfPlus, _selfSupMinus * _otherSupMinus) - max(_selfSupPlus * _otherInfMinus, _selfInfMinus * _otherSupPlus), \
                                    max(_selfSupPlus * _otherSupPlus, _selfInfMinus * _otherInfMinus) - max(_selfInfPlus * _otherSupMinus, _selfSupMinus * _otherInfPlus))

        else:
            return other * self

    def __truediv__(self, other):
        if isinstance(other, ClassicalArithmetic):
            assert not (0 in other), 'Нельзя делить на нуль содержащие интервалы!'

            other = ClassicalArithmetic(1/other.b, 1/other.a)
            _selfInfPlus, _selfSupPlus = max(self._a, 0), max(self._b, 0)
            _selfInfMinus, _selfSupMinus = _selfInfPlus - self._a, _selfSupPlus - self._b

            _otherInfPlus, _otherSupPlus = max(other.a, 0), max(other.b, 0)
            _otherInfMinus, _otherSupMinus = _otherInfPlus - other.a, _otherSupPlus - other.b

            return KaucherArithmetic(_selfInfPlus*_otherInfPlus + _selfSupMinus*_otherSupMinus - _selfSupPlus*_otherInfMinus - _selfInfMinus*_otherSupPlus, \
                                    _selfSupPlus*_otherSupPlus + _selfInfMinus*_otherInfMinus - _selfInfPlus*_otherSupMinus - _selfSupMinus*_otherInfPlus)

        if isinstance(other, KaucherArithmetic):
            assert not (0 in other), 'Нельзя делить на нуль содержащие интервалы!'
            return self.__mul__(KaucherArithmetic(1/other.b, 1/other.a))

        else:
            return 1/other * self

    def __pow__(self, other):
        if isinstance(other, (int, np.int32, np.int64)) and other >= 0:
            inf, sup = self._a**other, self._b**other
            if other % 2 == 0 and 0 in self:
                return KaucherArithmetic(0, max(inf, sup))
            else:
                return KaucherArithmetic(inf, sup)
        elif self >= 0:
            return np.exp(other * np.log(self))
        else:
            raise ValueError('Если основание содержит отрицательные числа, то степень может быть только натуральным числом.')

    # Инверсные арифметические операторы
    def __radd__(self, other):
        if isinstance(other, single_type):
            return KaucherArithmetic(self._a + other, self._b + other)
        else:
            return ArrayInterval(np.vectorize(lambda o: o + self)(other))

    def __rsub__(self, other):
        if isinstance(other, single_type):
            return KaucherArithmetic(other - self._b, other - self._a)
        else:
            return ArrayInterval(np.vectorize(lambda o: o - self)(other))

    def __rmul__(self, other):
        def fmul(x, y):
            _selfInfPlus, _selfSupPlus = max(y.a, 0), max(y.b, 0)
            _selfInfMinus, _selfSupMinus = _selfInfPlus - y.a, _selfSupPlus - y.b

            _otherInfPlus = max(x, 0)
            _otherSupPlus = _otherInfPlus
            _otherInfMinus, _otherSupMinus = _otherInfPlus - x, _otherSupPlus - x

            return KaucherArithmetic(_selfInfPlus*_otherInfPlus + _selfSupMinus*_otherSupMinus - _selfSupPlus*_otherInfMinus - _selfInfMinus*_otherSupPlus, \
                        _selfSupPlus*_otherSupPlus + _selfInfMinus*_otherInfMinus - _selfInfPlus*_otherSupMinus - _selfSupMinus*_otherInfPlus)

        if isinstance(other, single_type):
            return fmul(other, self)
        else:
            return ArrayInterval(np.vectorize(lambda o: o * self)(other))

    def __rtruediv__(self, other):
        return other * KaucherArithmetic(1/self._b, 1/self._a)

    #  Арифметические операторы присваивания
    def __iadd__(self, other):
        other = Interval(other)
        if isinstance(other, ARITHMETICS):
            self._a, self._b = self._a + other.a, self._b + other.b
            return self
        else:
            raise SyntaxError("It is not possible to change the type of the variable.")

    def __isub__(self, other):
        other = Interval(other)
        if isinstance(other, ARITHMETICS):
            self._a, self._b = self._a - other.b,  self._b - other.a
            return self
        else:
            raise SyntaxError("It is not possible to change the type of the variable.")

    def __imul__(self, other):
        other = Interval(other)
        if isinstance(other, ARITHMETICS):
            _selfInfPlus, _selfSupPlus = max(self._a, 0), max(self._b, 0)
            _selfInfMinus, _selfSupMinus = _selfInfPlus - self._a, _selfSupPlus - self._b

            _otherInfPlus, _otherSupPlus = max(other.a, 0), max(other.b, 0)
            _otherInfMinus, _otherSupMinus = _otherInfPlus - other.a, _otherSupPlus - other.b

            self._a = max(_selfInfPlus * _otherInfPlus, _selfSupMinus * _otherSupMinus) - max(_selfSupPlus * _otherInfMinus, _selfInfMinus * _otherSupPlus)
            self._b = max(_selfSupPlus * _otherSupPlus, _selfInfMinus * _otherInfMinus) - max(_selfInfPlus * _otherSupMinus, _selfSupMinus * _otherInfPlus)
            return self

        else:
            raise SyntaxError("It is not possible to change the type of the variable.")

    def __itruediv__(self, other):
        other = Interval(other)
        if isinstance(other, ClassicalArithmetic):
            assert (not 0 in other), 'Нельзя делить на нуль содержащие интервалы!'

            other = ClassicalArithmetic(1/other.b, 1/other.a)
            _selfInfPlus, _selfSupPlus = max(self._a, 0), max(self._b, 0)
            _selfInfMinus, _selfSupMinus = _selfInfPlus - self._a, _selfSupPlus - self._b

            _otherInfPlus, _otherSupPlus = max(other.a, 0), max(other.b, 0)
            _otherInfMinus, _otherSupMinus = _otherInfPlus - other.a, _otherSupPlus - other.b

            self._a = _selfInfPlus*_otherInfPlus + _selfSupMinus*_otherSupMinus - _selfSupPlus*_otherInfMinus - _selfInfMinus*_otherSupPlus
            self._b = _selfSupPlus*_otherSupPlus + _selfInfMinus*_otherInfMinus - _selfInfPlus*_otherSupMinus - _selfSupMinus*_otherInfPlus
            return self

        if isinstance(other, KaucherArithmetic):
            assert not (0 in other), 'Нельзя делить на нуль содержащие интервалы!'

            div = self.__mul__(KaucherArithmetic(1/other.b, 1/other.a))
            self._a, self._b = div._a, div._b
            return self

        else:
            raise SyntaxError("It is not possible to change the type of the variable.")


class ArrayInterval:

    def __init__(self, intervals):
        self._data = intervals

        self._shape = self._data.shape
        self._ndim = self._data.ndim
        self._ranges = [range(0, self._shape[k]) for k in range(self._ndim)]

    def __repr__(self):
        return 'Interval' + self._data.__repr__()[5:-15] + ')'

    @property
    def a(self):
        return np.vectorize(lambda el: el.a)(self._data)

    @property
    def b(self):
        return np.vectorize(lambda el: el.b)(self._data)

    @property
    def inf(self):
        return np.vectorize(lambda el: el.a)(self._data)

    @property
    def sup(self):
        return np.vectorize(lambda el: el.b)(self._data)

    @property
    def data(self):
        return self._data

    @property
    def shape(self):
        return self._shape

    @property
    def ndim(self):
        return self._ndim

    @property
    def ranges(self):
        return self._ranges

    def to_float(self):
        return ArrayInterval(np.vectorize(lambda el: el.to_float())(self._data))

    @property
    def rad(self):
        return np.vectorize(lambda el: el.rad)(self._data)

    @property
    def mid(self):
        return np.vectorize(lambda el: el.mid)(self._data)

    @property
    def wid(self):
        return np.vectorize(lambda el: el.wid)(self._data)

    @property
    def mig(self):
        return np.vectorize(lambda el: el.mig)(self._data)

    @property
    def mag(self):
        return np.vectorize(lambda el: el.mag)(self._data)

    @property
    def dual(self):
        return ArrayInterval(np.vectorize(lambda el: el.dual)(self._data))

    @property
    def pro(self):
        return ArrayInterval(np.vectorize(lambda el: el.pro)(self._data))

    @property
    def opp(self):
        return ArrayInterval(np.vectorize(lambda el: el.opp)(self._data))

    @property
    def inv(self):
        return ArrayInterval(np.vectorize(lambda el: el.inv)(self._data))

    @property
    def vertex(self):
        inf, sup = self.a, self.b
        if self._ndim == 1:
            n = self._shape[0]
            result = np.zeros((2**n, n))
            k = 0
            for ends in itertools.product([0, 1], repeat=n):
                ends = np.array(ends)
                result[k][ends == 0] = inf[ends == 0]
                result[k][ends == 1] = sup[ends == 1]
                k += 1

        elif self._ndim == 2:
            n, m = self._shape
            result = np.zeros((2**(n*m), n, m))
            k = 0
            for ends in itertools.product(itertools.product([0, 1], repeat=m), repeat=n):
                ends = np.array(ends)
                result[k][ends == 0] = inf[ends == 0]
                result[k][ends == 1] = sup[ends == 1]
                k += 1
        else:
            raise Exception('Функция предусмотрена не более чем для двумерных массивов.')

        return result

    @property
    def copy(self):
        return ArrayInterval(np.copy(self._data))

    def __deepcopy__(self, memo):
        return ArrayInterval(np.copy(self._data))

    def __array__(self):
        return self._data

    @property
    def T(self):
        return ArrayInterval(self._data.T)

    def reshape(self, shape):
        return ArrayInterval(self._data.reshape(shape))

    def __neg__(self):
        return ArrayInterval(-self._data)

    def __reversed__(self):
        return iter(self.storage[::-1])

    def __abs__(self):
        return np.vectorize(lambda el: abs(el))(self._data)

    def __contains__(self, other):
        if isinstance(other, single_type) or isinstance(other, (ClassicalArithmetic, KaucherArithmetic)):
            for index in itertools.product(*self._ranges):
                contain = other in self._data[index]
                if contain:
                    return contain
        else:
            for index in itertools.product(*self._ranges):
                contain = other[index] in self._data[index]
                if contain:
                    return contain
        return contain

    def __it__(self, other):
        if isinstance(other, ArrayInterval):
            return self._data < other._data
        else:
            return self._data < other

    def __le__(self, other):
        if isinstance(other, ArrayInterval):
            return self._data <= other._data
        else:
            return self._data <= other

    def __eq__(self, other):
        if isinstance(other, ArrayInterval):
            return self._data == other._data
        else:
            return self._data == other

    def __ne__(self, other):
        if isinstance(other, ArrayInterval):
            return self._data != other._data
        else:
            return self._data != other

    def __gt__(self, other):
        if isinstance(other, ArrayInterval):
            return self._data > other._data
        else:
            return self._data > other

    def __ge__(self, other):
        if isinstance(other, ArrayInterval):
            return self._data >= other._data
        else:
            return self._data >= other

    #     Эмуляция коллекций
    def __len__(self):
        return self._data.__len__()

    def __getitem__(self, key):
        result = self._data[key]
        if isinstance(result, INTERVAL_CLASSES):
            return result
        else:
            return ArrayInterval(result)

    def __setitem__(self, key, value):
        if isinstance(value, ARITHMETICS):
            self._data[key] = value
        elif isinstance(value, ArrayInterval):
            self._data[key] = value._data
        else:
            if isinstance(value, single_type):
                self._data[key] = SingleInterval(value, value)
            else:
                self._data[key] = ArrayInterval(np.vectorize(lambda v: SingleInterval(v, v))(value))._data

    def __delitem__(self, key):
        if isinstance(key, int):
            self._data = np.delete(self._data, key, axis=0)
            self._shape = self._data.shape
            self._ndim = self._data.ndim
            self._ranges = [range(0, self._shape[k]) for k in range(self._ndim)]
        elif isinstance(key, (slice, tuple, list, np.ndarray)):
            self._data = np.delete(self._data, key[-1], axis=len(key)-1)
            self._shape = self._data.shape
            self._ndim = self._data.ndim
            self._ranges = [range(0, self._shape[k]) for k in range(self._ndim)]
        else:
            msg = 'Indices must be integers / slice / tuple / list / np.ndarray'
            raise TypeError(msg)

    def __add__(self, other):
        if isinstance(other, ARITHMETICS):
            return self._data + other
        elif isinstance(other, ArrayInterval):
            return ArrayInterval(self._data + other._data)
        else:
            return ArrayInterval(self._data + other)

    def __sub__(self, other):
        if isinstance(other, ARITHMETICS):
            return self._data - other
        elif isinstance(other, ArrayInterval):
            return ArrayInterval(self._data - other._data)
        else:
            return ArrayInterval(self._data - other)

    def __mul__(self, other):
        if isinstance(other, ARITHMETICS):
            return self._data * other
        elif isinstance(other, ArrayInterval):
            return ArrayInterval(self._data * other._data)
        else:
            return ArrayInterval(self._data * other)

    def __truediv__(self, other):
        if isinstance(other, ARITHMETICS):
            return self._data / other
        elif isinstance(other, ArrayInterval):
            return ArrayInterval(self._data / other._data)
        else:
            return ArrayInterval(self._data / other)

    def __pow__(self, other):
        return ArrayInterval(np.vectorize(pow)(self._data, other))

    def __matmul__(self, other):
        matmul = self._data @ other
        if isinstance(matmul, INTERVAL_CLASSES):
            return matmul
        else:
            return ArrayInterval(matmul)

    def __radd__(self, other):
        return ArrayInterval(self._data + other)

    def __rsub__(self, other):
        return ArrayInterval(other - self._data)

    def __rmul__(self, other):
        return ArrayInterval(self._data * other)

    def __rtruediv__(self, other):
        return ArrayInterval(1 / self._data).__rmul__(other)

    def __rmatmul__(self, other):
        matmul = np.array(other) @ self._data
        if isinstance(matmul, INTERVAL_CLASSES):
            return matmul
        else:
            return ArrayInterval(matmul)

    def __iadd__(self, other):
        self._data = np.array(self._data + other)
        return self

    def __isub__(self, other):
        self._data = np.array(self._data - other)
        return self

    def __imul__(self, other):
        self._data = np.array(self._data * other)
        return self

    def __itruediv__(self, other):
        self._data = np.array(self._data / other)
        return self

    def __array_ufunc__(self, *args):
        if args[0].__name__ in ['add']:
            return ArrayInterval(self._data + args[2])

        elif args[0].__name__ in ['subtract']:
            return ArrayInterval(args[2] - self._data)

        elif args[0].__name__ in ['multiply']:
            return ArrayInterval(args[2] * self._data)

        elif args[0].__name__ in ['true_divide']:
            return ArrayInterval(args[2] / self._data)

        elif args[0].__name__ in ['matmul']:
            matmul = args[2] @ self._data
            if isinstance(matmul, INTERVAL_CLASSES):
                return matmul
            else:
                return ArrayInterval(matmul)

        elif args[0].__name__ in ['sqrt']:
            return ArrayInterval(np.vectorize(np.sqrt)(self._data))

        elif args[0].__name__ in ['exp']:
            return ArrayInterval(np.vectorize(np.exp)(self._data))

        elif args[0].__name__ in ['log']:
            return ArrayInterval(np.vectorize(np.log)(self._data))

        elif args[0].__name__ in ['sin']:
            return ArrayInterval(np.vectorize(np.sin)(self._data))

        elif args[0].__name__ in ['cos']:
            return ArrayInterval(np.vectorize(np.cos)(self._data))

        else:
            raise Exception("Расчёт функции {} не предусмотрен!".format(args[0].__name__))


class precision:
    increasedPrecisionQ = True
    mp.dps = 36

    def dps(_dps):
        mp.dps = _dps


def SingleInterval(left, right, sortQ=True, midRadQ=False):
    if precision.increasedPrecisionQ:
        left, right = mpf(np.str(left)), mpf(np.str(right))
    else:
        left, right = np.float64(left), np.float64(right)

    if midRadQ:
        assert right >= 0, "The radius of the interval cannot be negative."
        left, right = left - right, left + right

    if sortQ:
        if left > right:
            return ClassicalArithmetic(right, left)
        else:
            return ClassicalArithmetic(left, right)
    else:
        if left <= right:
            return ClassicalArithmetic(left, right)
        else:
            return KaucherArithmetic(left, right)

single_type = (int, float, np.int, np.float, np.int32, np.float32, np.int64, np.float64, mpf)
ARITHMETICS = (ClassicalArithmetic, KaucherArithmetic)
INTERVAL_CLASSES = (ClassicalArithmetic, KaucherArithmetic, ArrayInterval)


def Interval(*args, sortQ=True, midRadQ=False):

    n = len(args)
    assert n <= 2, "There cannot be more than two main arguments."

    if n == 2:
        if isinstance(args[0], single_type):
            return SingleInterval(args[0], args[1], sortQ=sortQ, midRadQ=midRadQ)
        left, right = np.array(args[0]), np.array(args[1])
        data = np.stack((left, right), axis=left.ndim)
    else:
        if isinstance(args[0], INTERVAL_CLASSES):
            return args[0]
        elif isinstance(args[0], single_type):
            return SingleInterval(args[0], args[0], sortQ=sortQ, midRadQ=midRadQ)
        data = np.array(args[0])

    return ArrayInterval(np.vectorize(lambda l, r: SingleInterval(l, r, sortQ=sortQ, midRadQ=midRadQ))(data[..., 0], data[..., 1]))



import numpy as np
import itertools
from collections.abc import Sequence

from .RealInterval import ClassicalArithmetic, KaucherArithmetic, ArrayInterval, Interval, INTERVAL_CLASSES, single_type, ARITHMETICS

infinity = float('inf')
nan = float('nan')


def get_shape(lst, shape=()):
    """
    Возвращает форму вложенных списков аналогично форме numpy.

    Parameters:
                lst: list
                    Вложенный список.

                shape: tuple
                    Форма до текущей глубины рекурсии.

    Returns:
                out: tuple
                    Форма текущей глубины. В конце это будет полная глубина.
    """

    if isinstance(lst, single_type) or (len(lst) == 1 and isinstance(lst, INTERVAL_CLASSES)):
        return shape
    elif isinstance(lst, np.ndarray):
        return lst.shape

    if isinstance(lst[0], Sequence):
        l = len(lst[0])
        if not all(len(item) == l for item in lst):
            msg = 'Не все списки имеют одинаковую длину!'
            raise ValueError(msg)

    shape += (len(lst), )

    # рекусрсия
    shape = get_shape(lst[0], shape)

    return shape


def asinterval(a):
    """
    Преобразование входных данных в массив типа Interval.

    Parameters:
                a: array_like
                    Входные данные, в любой форме, которые могут быть преобразованы
                    в массив интервалов.
                    Это включает в себя int, float, list и ndarrays.

    Returns:
                out: Interval
                    Преобразование не выполняется, если входные данные уже являются
                    типом Interval.
                    Если a - int, float, list или ndarrays, то возвращается
                    базовый класс Interval.
    """
    if isinstance(a, INTERVAL_CLASSES):
        return a

    elif isinstance(a, single_type):
        return Interval(a, a, sortQ=False)

    elif isinstance(a, (list, np.ndarray)):
        a = np.asarray(a)
        shape = get_shape(a)

        result = zeros(shape)
        for index in itertools.product(*result.ranges):
            if isinstance(a[index], INTERVAL_CLASSES):
                result[index] = a[index]
            else:
                result[index] = Interval(a[index], a[index], sortQ=False)
        return result

    else:
        msg = 'Входные данные неприводимы к интервальному массиву!'
        raise TypeError(msg)


def intersection(x, y):
    """
    Покомпонентное пересечение двух интервальных массивов.

    Parameters:
                A, B: Interval
                    В случае, если операнды не являются интервальным типом, то
                    они преобразуются функцией asinterval.

    Returns:
                out: Interval
                    Возвращается массив пересечённых интервалов.
                    Если некоторые интервалы не пересекаются, то на их месте
                    выводится интервал Interval(None, None).
    """

    def intersect(x, y):
        sup = y.a if x.a < y.a else x.a
        inf = x.b if x.b < y.b else y.b
        if sup - inf <= 1e-15:
            return Interval(sup, inf, sortQ=True)
        else:
            return Interval(nan, nan)

    if isinstance(x, ArrayInterval) and isinstance(y, ArrayInterval):
        assert x.shape == y.shape, 'Не совпадают размерности входных массивов!'
        return ArrayInterval(np.vectorize(intersect)(x.data, y.data))
    elif isinstance(x, ARITHMETICS) and isinstance(y, ArrayInterval):
        return ArrayInterval(np.vectorize(intersect)(x, y.data))
    elif isinstance(x, ArrayInterval) and isinstance(y, ARITHMETICS):
        return ArrayInterval(np.vectorize(intersect)(x.data, y))
    else:
        return intersect(x, y)

def dist(x, y, order=infinity):
    """
    Метрика в интервальных пространствах.

    Parameters:
                a, b: Interval
                    Интервалы между которыми необходимо рассчитать dist.
                    В случае многомерности операндов вычисляется мультиметрика.

                order: int
                    Задаются различные метрики. По умолчанию используется
                    Чебышёвское расстояние.

    Returns:
                out: float
                    Возвращается расстояние между входными операндами.
    """

    def cheb(x, y):
        return np.maximum(abs(x.a - y.a), abs(x.b - y.b))

    if order == infinity:
        return np.amax(cheb(x, y))
    elif isinstance(order, int):
        return pow(np.sum(cheb(x, y) ** order), 1/order)
    else:
        raise Exception('Не верно задан порядок нормы order.')


def zeros(shape):
    """Функция создаёт массив размерности shape."""
    return Interval(np.zeros(shape, dtype='float64'),
                    np.zeros(shape, dtype='float64'), sortQ=False)


def diag(mat):
    """Функция возвращает диагональные элементы матрицы."""
    mat = asinterval(mat)

    if mat.ndim != 2:
        raise Exception('Входной массив не является двумерным!')
    n = min(mat.shape)
    result = zeros(n)
    for k in range(n):
        result[k] = mat[k, k]
    return result


def eye(n):
    A = zeros((n, n))
    for k in range(n):
        A[k, k] = Interval(1, 1, sortQ=False)
    return A


def Neumeier(n, theta, infb=None, supb=None):
    interval = [None, None]
    if infb is None:
        interval[0] = -1
    else:
        interval[0] = infb
    if supb is None:
        interval[1] = 1
    else:
        interval[1] = supb
    b = Interval([interval]*n)

    A = zeros((n,n)) + Interval(0, 2, sortQ=False)
    for k in range(n):
        A[k, k] = Interval(theta, theta, sortQ=False)

    return A, b


def Shary(n, N=None, alpha=0.23, beta=0.35):
    if N is None:
        N = n
    else:
        assert N>=n-1, "Параметры заданы неверно!"

    if alpha < 0 or beta < alpha or beta > 1:
        raise Exception('Параметры заданы неверно!')

    A = zeros((n,n)) + Interval(alpha-1, 1-beta, sortQ=False)
    b = zeros(n) + Interval(1-n, n-1, sortQ=False)

    for k in range(n):
        A[k, k] = Interval(n-1, N, sortQ=False)

    return A, b


def create_data(x=None, N=3, model=None):

    def T(n, x):
        if n == 0:
            return 1
        elif n == 1:
            return x
        else:
            return 2 * x * T(n-1, x) - T(n-2, x)

    if x is None:
        x = uniform(-1, 1, 10)

    if model is None:
        model = lambda x: x ** 2

    b = model(x)
    A = []
    for k in range(len(x)):
        A.append([T(l, x[k]) for l in range(N)])
    A = asinterval(A)

    return A, b


def randint(inf, sup, shape=1):
    return Interval(np.random.randint(inf, sup, shape),
                    np.random.randint(inf, sup, shape))


def uniform(inf, sup, shape=1):
    return Interval(np.random.uniform(inf, sup, shape),
                    np.random.uniform(inf, sup, shape))


def normal(mu, sigma, shape=1):
    return Interval(np.random.normal(mu, sigma, shape),
                    np.random.normal(mu, sigma, shape))


def dot(a, b, aspotQ=False, bspotQ=False):
    if aspotQ:
        midb = b.mid
        radb = b.rad

        tmp1 = np.dot(a, midb)
        tmp2 = np.dot(a.mag, radb)

        return Interval(tmp1 - tmp2, tmp1 + tmp2, sortQ=False)

    elif bspotQ:
        mida = a.mid
        rada = a.rad

        tmp1 = np.dot(mida, b)
        tmp2 = np.dot(rada, b.mag)

        return Interval(tmp1 - tmp2, tmp1 + tmp2, sortQ=False)

    else:
        return a @ b

def isnan(x):
    def _isnan(x):
        isnanQ = np.isnan(float(x.a)) or np.isnan(float(x.b))
        return isnanQ
    if isinstance(x, ARITHMETICS):
        return _isnan(x)
    else:
        return np.vectorize(lambda xx: _isnan(xx))(x.data)


subset = lambda a, b: np.array(((a.a >= b.a) & (a.b <= b.b)), dtype=np.bool).all()
superset = lambda a, b: subset(b, a)

proper_subset = lambda a, b: np.array(((a.a > b.a) & (a.b < b.b)), dtype=np.bool).all()
proper_superset = lambda a, b: proper_subset(b, a)

contain = lambda a, b: np.array(((a.a >= b.a) & (a.b <= b.b)), dtype=np.bool)
supercontain = lambda a, b: subset(b, a)

def sqrt(x):
    """Interval enclosure of the square root intrinsic over an interval."""
    return np.sqrt(x)

def exp(x):
    """Interval enclosure of the exponential intrinsic over an interval."""
    return np.exp(x)

def log(x):
    """Interval enclosure of the natural logarithm intrinsic over an interval."""
    return np.log(x)

def sin(x):
    """Interval enclosure of the sin intrinsic over an interval."""
    return np.sin(x)

def cos(x):
    """Interval enclosure of the cos intrinsic over an interval."""
    return np.sin(x)


def sgn(x):
    def _sgn(x):
        if x.b < 0:
            return Interval(-1, -1, sortQ=False)
        elif x.a < 0 and x.b == 0:
            return Interval(-1, 0, sortQ=False)
        elif x.a < 0 and 0 < x.b:
            return Interval(-1, 1, sortQ=False)
        elif x.a == 0 and x.b == 0:
            return Interval(0, 0, sortQ=False)
        elif x.a == 0 and 0 < x.b:
            return Interval(0, 1, sortQ=False)
        else:
            return Interval(1, 1, sortQ=False)

    if isinstance(x, ARITHMETICS):
        return _sgn(x)
    else:
        return asinterval(np.vectorize(_sgn)(x.data))



import numpy as np
from mpmath import mpf

from intvalpy.RealInterval import Interval, precision
from intvalpy.utils import asinterval, infinity


def __tolsolvty(infA, supA, infb, supb, weight=None, x0=None,
                tol_f=1e-12, tol_x=1e-12, tol_g=1e-12, maxiter=2000):

    nsims = 30
    alpha = 2.3
    hs = 1
    nh = 3
    q1 = 0.9
    q2 = 1.1

    m, n = infA.shape
    if weight is None:
        if precision.increasedPrecisionQ:
            weight = np.array([mpf('1')]*m, dtype=np.object)
        else:
            weight = np.ones(m)

    Ac = 0.5 * (infA + supA)
    Ar = 0.5 * (supA - infA)
    bc = 0.5 * (infb + supb)
    br = 0.5 * (supb - infb)

    sv = np.linalg.svd(np.array(Ac, dtype=np.float64), compute_uv=False)
    if precision.increasedPrecisionQ:
        sv = np.array([mpf(np.str(el)) for el in sv], dtype=np.object)
    minsv, maxsv = np.min(sv), np.max(sv)

    tt = None

    def calcfg(x):
        index = x >= 0

        Ac_x = Ac @ x
        Ar_absx = Ar @ np.abs(x)
        infs = bc - (Ac_x + Ar_absx)
        sups = bc - (Ac_x - Ar_absx)

        tt = weight * (br - np.maximum(np.abs(infs), np.abs(sups)))
        mc = np.argmin(tt)

        if -infs[mc] <= sups[mc]:
            dd = weight[mc] * (infA[mc] * index + supA[mc] * (~index))
        else:
            dd = -weight[mc] * (supA[mc] * index + infA[mc] * (~index))

        return tt[mc], dd, tt

    if x0 is None:
        if (minsv != 0 and maxsv/minsv < 1e15):
            x = np.linalg.lstsq(np.array(Ac, dtype=np.float64), np.array(bc, dtype=np.float64), rcond=-1)[0]
            if precision.increasedPrecisionQ:
                sv = np.array([mpf(np.str(xx)) for xx in x], dtype=np.object)
        else:
            if precision.increasedPrecisionQ:
                x = np.array([mpf('0')]*n, dtype=np.object)
            else:
                x = np.zeros(n)
    else:
        x = np.copy(x0)

    B = np.eye(n)
    vf = np.zeros(nsims) + infinity
    w = 1./alpha - 1

    f, g0, _ = calcfg(x)
    ff = f;  xx = x
    cal = 1;  ncals = 1
    ccode = False

    for nit in range(int(maxiter)):
        vf[nsims-1] = ff
        if np.linalg.norm(g0) < tol_g:
            ccode = True
            break

        g1 = B.T @ g0
        g = B @ (g1/np.linalg.norm(g1))
        normg = np.linalg.norm(g)

        r = 1
        cal = 0
        deltax = 0

        while r > 0 and cal <= 500:
            cal += 1
            x = x + hs * g
            deltax = deltax + hs * normg
            f, g1, tt = calcfg(x)

            if f > ff:
                ff = f
                xx = x

            if np.mod(cal, nh) == 0:
                hs = hs * q2
            r = g @ g1

        if cal > 500:
            ccode = False
            break

        if cal == 1:
            hs = hs * q1

        ncals = ncals + cal
        if deltax < tol_x:
            ccode = True
            break

        dg = B.T @ (g1 - g0)
        xi = dg / np.linalg.norm(dg)
        B = B + w * np.outer((B @ xi), xi)
        g0 = g1

        vf = np.roll(vf, 1)
        vf[0] = abs(ff - vf[0])

        if abs(ff) > 1:
            deltaf = np.sum(vf)/abs(ff)
        else:
            deltaf = np.sum(vf)

        if deltaf < tol_f:
            ccode = True
            break
        ccode = False

    return ccode, xx, ff, tt


def Uni(A, b, x=None, maxQ=False, x0=None, tol=1e-12, maxiter=2000):
    """
    When it is necessary to check an interval system of linear equations for its weak solvability
    you should use the Uni functionality. If maxQ=True, then the maximum of the functional is found,
    otherwise, the value at point x is calculated.

    To optimize it, the well-known Nelder-Mead method is used, which does not use gradients,
    since there is an absolute value in the function.


    Parameters:

        A: Interval
            The input interval matrix of ISLAE, which can be either square or rectangular.

        b: Interval
            The interval vector of the right part of the ISLAE.

        x: float, array_like, optional
            The point at which the recognizing functional is calculated. By default, x is equal to an array of zeros.

        maxQ: bool, optional
            If the parameter value is True, then the functional is maximized.

        x0: float, array_like, optional
            The initial guess for finding the global maximum.

        tol: float, optional
            Absolute error in xopt between iterations that is acceptable for convergence.

        maxiter: int, optional
            The maximum number of iterations.

    Returns:

        out: float, tuple
            The value of the recognizing functional at point x is returned.
            If maxQ=True, then a tuple is returned, where the first element is the correctness of the optimization completion,
            the second element is the optimum point, and the third element is the value of the function at this point.
    """

    from scipy.optimize import minimize, LinearConstraint

    midA = A.mid
    radA = A.rad
    br = b.rad
    bm = b.mid

    def __dot(x):
        tmp1 = np.dot(midA, x)
        tmp2 = np.dot(radA, abs(x))
        return Interval(tmp1 - tmp2, tmp1 + tmp2, sortQ=False)

    def __uni(x):
        return min(br - (bm - __dot(x)).mig)
    __minus_uni = lambda x: -__uni(x)

    if not maxQ:
        if x is None:
            x = np.zeros(A.shape[1])
        return __uni(x)
    else:
        if x0 is None:
            _, x0, _, _ = __tolsolvty(A.a, A.b, b.a, b.b,
                                      tol_f=1e-8, tol_x=1e-8, tol_g=1e-8)
        maximize = minimize(__minus_uni, x0, method='COBYLA', tol=tol, options={'maxiter': maxiter})

        return maximize.success, maximize.x, -maximize.fun


def Tol(A, b, x=None, maxQ=False, weight=None, x0=None, tol=1e-12, maxiter=2000):
    """
    When it is necessary to check the interval system of linear equations for its strong
    solvability you should use the Tol functionality. If maxQ=True, then the maximum
    of the functional is found, otherwise, the value at point x is calculated.
    To optimize it, a proven the tolsolvty program, which is suitable for solving practical problems.


    Parameters:

        A: Interval
            The input interval matrix of ISLAE, which can be either square or rectangular.

        b: Interval
            The interval vector of the right part of the ISLAE.

        x: float, array_like, optional
            The point at which the recognizing functional is calculated. By default, x is equal to an array of zeros.

        maxQ: bool, optional
            If the parameter value is True, then the functional is maximized.

        x0: float, array_like, optional
            The initial guess for finding the global maximum.

        tol: float, optional
            Absolute error in xopt between iterations that is acceptable for convergence.

        maxiter: int, optional
            The maximum number of iterations.

    Returns:

        out: float, tuple
            The value of the recognizing functional at point x is returned.
            If maxQ=True, then a tuple is returned, where the first element is the correctness of the optimization completion,
            the second element is the optimum point, and the third element is the value of the function at this point.
    """

    br = b.rad
    bm = b.mid
    __tol = lambda x: np.min(br - (bm - A @ x).mag)
    __minus_tol = lambda x: -__tol(x)

    if not maxQ:
        if x is None:
            x = np.zeros(A.shape[1])
        return __tol(x)
    else:
        ccode, x, fun, _ = __tolsolvty(A.a, A.b, b.a, b.b, weight=weight, x0=x0, maxiter=maxiter,
                                        tol_f=tol, tol_x=tol, tol_g=tol)
        return ccode, x, fun


def ive(A, b, N=40):
    """
    Вычисление меры вариабельности оценки параметров.

    Parameters:
                A: Interval
                    Матрица ИСЛАУ.

                b: Interval
                    Вектор правой части ИСЛАУ.

    Optional Parameters:
                N: int
                    Количество угловых матриц для которых вычисляется обусловленность.

    Returns:
                out: float
                    Возвращается мера вариабельности IVE.
    """

    success, _arg_max, _max = Tol(A, b, maxQ=True)
    if not success:
        print('Оптимизация функционала Tol завершена некорректно!')

    _inf = A.a
    _sup = A.b
    cond = infinity
    angle_A = np.zeros(A.shape, dtype='float64')
    for _ in range(N):
        for k in range(A.shape[0]):
            for l in range(A.shape[1]):
                angle_A[k, l] = np.random.choice([_inf[k, l], _sup[k, l]])
        tmp = np.linalg.cond(angle_A)
        cond = tmp if tmp<cond else cond

    return np.sqrt(A.shape[1]) * _max * cond * (np.linalg.norm(_arg_max, ord=2)/np.sqrt(sum(b.mag**2)))


def outliers(A, b, functional='uni', x0=None, tol=1e-12, maxiter=2000, method='standard deviations'):

    def interquartile(data):
        q25, q75 = np.percentile(data, 25), np.percentile(data, 75)
        iqr = q75 - q25
        cut_off = iqr * 1.5

        lower, upper = q25 - cut_off, q75 + cut_off
        return np.argwhere((data < lower) | (data > upper)).flatten()

    def standard_deviations(data):
        # Set upper and lower limit to 3 standard deviation
        std, mean = np.std(data), np.mean(data)
        cut_off = std * 3

        lower, upper = mean - cut_off, mean + cut_off
        return np.argwhere((data < lower) | (data > upper)).flatten()

    WorkListA = asinterval(A).copy
    WorkListb = asinterval(b).copy

    if functional == 'uni':
        _, x, _ = Uni(A, b, maxQ=True, x0=x0, tol=tol, maxiter=maxiter)
        tt = WorkListb.rad - (WorkListb.mid - WorkListA @ x).mig

    elif functional == 'tol':
        _, _, _, tt = __tolsolvty(WorkListA.a, WorkListA.b, WorkListb.a, WorkListb.b, weight=x0, maxiter=maxiter,
                                  tol_f=tol, tol_x=tol, tol_g=tol)
    else:
        Exception('Данный функционал не предусмотрен.')

    if method == 'standard deviations':
        outliers_index = standard_deviations(tt)
    elif method == 'interquartile':
        outliers_index = interquartile(tt)
    else:
        Exception('Данный метод не предусмотрен.')

    index = np.delete(np.arange(WorkListA.shape[0]), outliers_index)
    WorkListA = WorkListA[index]
    WorkListb = WorkListb[index]

    return WorkListA, WorkListb, outliers_index, tt



import numpy as np

from intvalpy.RealInterval import Interval
from intvalpy.utils import asinterval, zeros, dist, intersection, diag


def Gauss(A, b):
    """
    Метод Гаусса для решения ИСЛАУ.

    Parameters:
                A: Interval
                    Матрица ИСЛАУ.

                b: Interval
                    Вектор правой части ИСЛАУ.

    Returns:
                out: Interval
                    Возвращается интервальный вектор решений.
    """

    WorkListA = asinterval(A).copy
    WorkListb = asinterval(b).copy

    n, _ = WorkListA.shape
    mignitude = diag(A).mig
    _abs_A = A.mag
    for k in range(n):
        if mignitude[k] < sum(_abs_A[k]) - _abs_A[k, k]:
            raise Exception('Матрица А не является H матрицей!')

    r = zeros((n, n))
    x = zeros(n)

    for j in range(n-1):
        r[j+1:, j] = WorkListA[j+1:, j]/WorkListA[j, j]
        WorkListA[j+1:, j+1:] = WorkListA[j+1:, j+1:] - r[j+1:, j] * WorkListA[j, j+1:]
        WorkListb[j+1:] = WorkListb[j+1:] - r[j+1:, j] * WorkListb[j]

    for i in range(n-1, -1, -1):
        x[i] = (WorkListb[i] - sum(WorkListA[i, i:] * x[i:])) / WorkListA[i, i]
    return x


def Gauss_Seidel(A, b, x0=None, P=True, tol=1e-8, maxiter=10**3):
    """
    Итерационный метод Гаусса-Зейделя для решения ИСЛАУ.

    Parameters:
                A: Interval
                    Матрица ИСЛАУ.

                b: Interval
                    Вектор правой части ИСЛАУ.

    Optional Parameters:
                x0: Interval
                    Начальный брус, в котором ищется решение.

                P: Interval
                    Матрица предобуславливания.
                    В случае, если параметр не задан, то берётся обратное среднее.

                tol: float
                    Погрешность для остановки итерационного процесса.

                maxiter: int
                    Максимальное количество итераций.

    Returns:
                out: Interval
                    Возвращается интервальный вектор решений.
    """

    A = asinterval(A).copy
    b = asinterval(b).copy

    if A.shape == () or A.shape == (1, 1):
        if 0 in A:
            raise Exception('Диагональный элемент матрицы содержит нуль!')
        return b/A

    if P:
        P = np.linalg.inv(np.array(A.mid, dtype=np.float64))
        A = P @ A
        b = P @ b

    n, _ = A.shape
    mignitude = diag(A).mig
    _abs_A = A.mag
    for k in range(n):
        if mignitude[k] < sum(_abs_A[k]) - _abs_A[k, k]:
            raise Exception('Матрица А не является H матрицей!')

    error = float("inf")
    result = zeros(n)

    if x0 is None:
        pre_result = zeros(n) + Interval(-1000, 1000, sortQ=False)
    else:
        pre_result = x0.copy

    nit = 0
    while error >= tol and nit <= maxiter:
        for k in range(n):
            tmp = 0
            for l in range(n):
                if l != k:
                    tmp += A[k, l] * pre_result[l]
            result[k] = intersection(pre_result[k], (b[k]-tmp)/A[k, k])

            if float('-inf') in result[k]:
                raise Exception("Интервалы не пересекаются!")

        error = dist(result, pre_result)
        pre_result = result.copy
        nit += 1

    return result


def Subdiff(A, b, tol=1e-12, maxiter=500, tau=1, norm_min_val=1e-12):

    def superMatrix(A):
        Amid = A.mid
        index = Amid >= 0
        A_plus = np.zeros(A.shape)
        A_minus = np.zeros(A.shape)
        A_plus[index] = Amid[index]
        A_minus[~index] = Amid[~index]

        result = np.zeros((2*n, 2*m))
        result[:n, :m], result[:n, m:2*m] = A_plus, A_minus
        result[n:2*n, :m], result[n:2*n, m:2*m] = A_minus, A_plus
        return result


    def calcSubgrad(F, i, j, a, b):
        n = int(F.shape[0] / 2)

        if np.sign(a.a) * np.sign(a.b) > 0:
            k = 0 if np.sign(a.a) > 0 else 2
        else:
            k = 1 if a.a < a.b else 3

        if np.sign(b.a) * np.sign(b.b) > 0:
            m = 1 if np.sign(b.a) > 0 else 3
        else:
            m = 2 if b.a <= b.b else 4

        cause = 4*k + m
        if cause == 1:
            F[i, j] = a.a
            F[i + n, j + n] = a.b
        elif cause == 2:
            F[i, j] = a.b
            F[i + n, j + n] = a.b
        elif cause == 3:
            F[i, j] = a.b
            F[i + n, j + n] = a.a
        elif cause == 4:
            F[i, j] = a.a
            F[i + n, j + n] = a.a
        elif cause == 5:
            F[i, j + n] = a.a
            F[i + n, j + n] = a.b
        elif cause == 6:
            if a.a*b.b < a.b*b.a:
                F[i, j + n] = a.a
            else:
                F[i, j] = a.b

            if a.a*b.a > a.b*b.b:
                F[i + n, j] = a.a
            else:
                F[i + n, j + n] = a.b
        elif cause == 7:
            F[i, j] = a.b
            F[i + n, j] = a.a
        elif cause == 9:
            F[i, j + n] = a.a
            F[i + n, j] = a.b
        elif cause == 10:
            F[i, j + n] = a.a
            F[i + n, j] = a.a
        elif cause == 11:
            F[i, j + n] = a.b
            F[i + n, j] = a.a
        elif cause == 12:
            F[i, j + n] = a.b
            F[i + n, j] = a.b
        elif cause == 13:
            F[i, j] = a.a
            F[i + n, j] = a.b
        elif cause == 15:
            F[i, j + n] = a.b
            F[i + n, j + n] = a.a
        elif cause == 16:
            if a.a*b.a > a.b*b.b:
                F[i, j] = a.a
            else:
                F[i, j + n] = -a.b

            if a.a*b.b < a.b*b.a:
                F[i + n, j + n] = a.a
            else:
                F[i + n, j] = a.b

        return F

    n, m = A.shape

    assert n == m, "matrix is not square"
    assert m == b.shape[0], "mismatch of matrix and vector dimensions"

    F = superMatrix(A)
    xx = np.zeros(2*n)
    xx[:n], xx[n:2*n] = b.a, b.b

    xx = np.linalg.solve(F, xx)
    r = float('inf')
    q = 1
    nit = 0
    while nit <= maxiter and r / q > tol:
        r = 0
        x = np.copy(xx)
        F = np.zeros((2*n, 2*n))

        for i in range(n):
            s = Interval(0, 0)

            for j in range(n):
                g = A[i, j]
                h = Interval(x[j], x[j+n], sortQ=False)
                t = g * h
                s = s + t
                F = calcSubgrad(F, i, j, g, h)

            t = s + b[i].opp
            xx[i] = t.a
            xx[i + n] = t.b

            r = r + t.mag

        xx = np.linalg.solve(F, xx)
        xx = x - xx * tau

        q = np.linalg.norm(xx, 1)
        if q <= norm_min_val:
            q = 1

        nit += 1
    return Interval(xx[:n], xx[n:], sortQ=False)



import numpy as np

from intvalpy.linear import Gauss_Seidel, PSS
from intvalpy.utils import asinterval, intersection, dist, infinity, isnan
from intvalpy.RealInterval import ARITHMETICS


def HansenSengupta(func, J, x0, maxiter=2000, tol=1e-12):

    def HS(X, c):
        L = asinterval(J(X))

        if isinstance(L, ARITHMETICS):
            LAMBDA = 1/L.mid
            A = LAMBDA * L
            b = -LAMBDA * func(c)

            if 0 in A:
                raise Exception('Диагональный элемент матрицы содержит нуль!')
            else:
                GS = b/A
            return c + GS

        else:
            LAMBDA = np.linalg.inv(L.to_float().mid)
            A = LAMBDA @ L
            b = -LAMBDA @ func(c)

            pss = PSS(A, b, tol=tol)
            return c + pss

    if not (0 in func(x0)):
        raise Exception('Брус не содержит решений!')

    result = x0
    pre_result = result.copy
    c = asinterval(result.mid)

    error = infinity
    nit = 0
    while nit < maxiter and error > tol:
        result = intersection(result, HS(result, c))
        if isnan(result).any():
            return result
        c = asinterval(result.mid)
        error = dist(result, pre_result)
        pre_result = result.copy
        nit += 1
    return result


def Krawczyk(func, J, x0, maxiter=2000, tol=1e-12):

    def K(X, c):
        L = asinterval(J(X))

        if isinstance(L, ARITHMETICS):
            LAMBDA = 1/L.mid

            B = 1 - LAMBDA * L
            return c - LAMBDA * asinterval(func(c)) + B * (X-c)

        else:
            n, m = L.shape
            LAMBDA = np.linalg.inv(L.to_float().mid)

            B = np.eye(n) - LAMBDA @ L
            w, _ = np.linalg.eigh(B.to_float().mag)
            return c - LAMBDA @ func(c) + B @ (X-c)

    if not (0 in func(x0)):
        raise Exception('Брус не содержит решений!')

    result = x0
    pre_result = result.copy
    c = asinterval(result.mid)

    error = infinity
    nit = 0
    while nit <= maxiter and error > tol:
        result = intersection(result, K(result, c))
        if isnan(result).any():
            return result
        c = asinterval(result.mid)
        error = dist(result, pre_result)
        pre_result = result.copy
        nit += 1

    return result



import unittest

import numpy as np
from intvalpy import Interval, precision

precision.increasedPrecisionQ = False

class TestArithmeticsOperation(unittest.TestCase):

    def test_add(self):
        assert Interval(1, 2) + 2                                                == Interval(3, 4)
        assert -1 + Interval(-2, 1)                                              == Interval(-3, 0)
        assert Interval(1, 2) + Interval(-2, 3)                                  == Interval(-1, 5)
        assert Interval(1, 2) + Interval(3, -2, sortQ=False)                     == Interval(4, 0, sortQ=False)

        assert Interval(2, 1, sortQ=False) + 2                                   == Interval(4, 3, sortQ=False)
        assert -1 + Interval(1, -2, sortQ=False)                                 == Interval(0, -3, sortQ=False)
        assert Interval(3, -2, sortQ=False) + Interval(-1, 2)                    == Interval(2, 0, sortQ=False)
        assert Interval(3, -2, sortQ=False) + Interval(2, -1, sortQ=False)       == Interval(5, -3, sortQ=False)

    def test_sub(self):
        assert Interval(1, 2) - 2                                                == Interval(-1, 0)
        assert -1 - Interval(-2, 1)                                              == Interval(-2, 1)
        assert Interval(1, 2) - Interval(-2, 3)                                  == Interval(-2, 4)
        assert Interval(1, 2) - Interval(3, -2, sortQ=False)                     == Interval(3, -1, sortQ=False)

        assert Interval(2, 1, sortQ=False) - 2                                   == Interval(0, -1, sortQ=False)
        assert -1 - Interval(1, -2, sortQ=False)                                 == Interval(1, -2, sortQ=False)
        assert Interval(3, -2, sortQ=False) - Interval(-1, 2)                    == Interval(1, -1, sortQ=False)
        assert Interval(3, -2, sortQ=False) - Interval(2, -1, sortQ=False)       == Interval(4, -4, sortQ=False)

    def test_mul(self):
        assert Interval(1, 2) * 2                                                == Interval(2, 4)
        assert -1 * Interval(-2, 1)                                              == Interval(-1, 2)
        assert Interval(1, 2) * Interval(-2, 3)                                  == Interval(-4, 6)
        assert Interval(1, 2) * Interval(3, -2, sortQ=False)                     == Interval(3, -2, sortQ=False)

        assert Interval(2, 1, sortQ=False) * 2                                   == Interval(4, 2, sortQ=False)
        assert -1 * Interval(1, -2, sortQ=False)                                 == Interval(2, -1, sortQ=False)
        assert Interval(3, -2, sortQ=False) * Interval(-1, 2)                    == Interval(0, 0, sortQ=False)
        assert Interval(3, -2, sortQ=False) * Interval(2, -1, sortQ=False)       == Interval(6, -4, sortQ=False)

    def test_div(self):
        assert Interval(1, 2) / 2                                                == Interval(0.5, 1)
        assert -1 / Interval(-2, -1)                                             == Interval(0.5, 1)
        assert Interval(1, 2) / Interval(2, 5)                                   == Interval(0.2, 1)
        assert Interval(1, 2) / Interval(4, 2, sortQ=False)                      == Interval(0.5, 0.5, sortQ=False)

        assert Interval(2, 1, sortQ=False) / 2                                   == Interval(1, 0.5, sortQ=False)
        assert -1 / Interval(-1, -2, sortQ=False)                                == Interval(1, 0.5, sortQ=False)
        assert Interval(3, -2, sortQ=False) / Interval(-1, -2)                   == Interval(1, -1.5, sortQ=False)
        assert Interval(3, -2, sortQ=False) / Interval(2, 1, sortQ=False)        == Interval(3, -2, sortQ=False)

    # def test_pow(self):
    #     assert Interval(-2, 3) ** 2                                              == Interval(0, 9)
    #     assert Interval(2, 3) ** 2                                               == Interval(4, 9)
    #     assert Interval(-3, -2) ** 2                                             == Interval(4, 9)
    #
    #     assert Interval(-2, 3) ** 3                                              == Interval(-8, 27)
    #     assert Interval(2, 3) ** 3                                               == Interval(8, 27)
    #     assert Interval(-3, -2) ** 3                                             == Interval(-27, -8)
    #
    #     assert Interval(3, -2, sortQ=False) ** 2                                 == Interval(9, 4, sortQ=False)     # уточнить
    #     assert Interval(3, 2, sortQ=False) ** 2                                  == Interval(9, 4, sortQ=False)
    #     assert Interval(-2, -3, sortQ=False) ** 2                                == Interval(4, 9, sortQ=False)
    #
    #     assert Interval(3, -2, sortQ=False) ** 3                                 == Interval(27, -8, sortQ=False)
    #     assert Interval(3, 2, sortQ=False) ** 3                                  == Interval(27, 8, sortQ=False)
    #     assert Interval(-2, -3, sortQ=False) ** 3                                == Interval(-8, -27, sortQ=False)

    def test_matmul(self):
        A = Interval([
            [[2, 4], [-2, 1]],
            [[-1, 2], [2, 4]]
        ])
        x = Interval([[-2, 2], [1, 2]])

        B = Interval([
            [[1, 2], [-7, 6]],
            [[-7, 6], [1, 2]]
        ])
        y = Interval([[-1, 1], [-2, 0]])


        assert (A @ B                                                            == Interval([[[-10, 22], [-32, 26]],
                                                                                              [[-30, 28], [-12, 20]]])).all()
        assert (A @ x                                                            == Interval([[-12, 10], [-2, 12]])).all()
        assert (x @ A                                                            == Interval([[-10, 12], [-2, 12]])).all()
        assert x @ y                                                             == Interval(-6, 2)


if __name__ == '__main__':
    unittest.main()



import numpy as np

from intvalpy.utils import asinterval, zeros, sgn
from intvalpy.RealInterval import Interval, INTERVAL_CLASSES

from bisect import bisect_left


class KeyWrapper:
    def __init__(self, iterable, key):
        self.it = iterable
        self.key = key

    def __getitem__(self, i):
        return self.key(self.it[i])

    def __len__(self):
        return len(self.it)


def __tolsolvty(func, grad, a, b, weight=None, x0=None,
                tol_f=1e-12, tol_x=1e-12, tol_g=1e-12, maxiter=2000, stepwise=float('inf')):

    nsims = 30
    alpha = 2.3
    hs = 1
    nh = 3
    q1 = 0.9
    q2 = 1.1

    if grad is None:
        raise Exception('Необходимо задать вектор градиент!')

    m = len(a)
    n = len(grad)
    if weight is None:
        weight = np.ones(m)

    bc = b.mid
    br = b.rad
    tt = None

    def calcfg(x):
        index = x >= 0
        infsup = bc - func(a, x)

        tt = weight * (br - np.maximum(np.abs(infsup.a), np.abs(infsup.b)))
        mc = np.argmin(tt)
        if n == 1:
            _grad = asinterval(grad[0](a[mc], x))
        else:
            _grad = asinterval([g(a[mc], x) for g in grad])

        if -infsup[mc].a <= infsup[mc].b:
            dd = weight[mc] * (_grad.a * index + _grad.b * (~index))
        else:
            dd = -weight[mc] * (_grad.b * index + _grad.a * (~index))

        return tt[mc], dd, tt


    if x0 is None:
        x = np.zeros(n)
    else:
        x = x0

    B = np.eye(n)
    vf = np.zeros(nsims) + float('inf')
    w = 1./alpha - 1

    f, g0, _ = calcfg(x)
    ff = f;  xx = x
    cal = 1;  ncals = 1
    ccode = False

    for nit in range(1, int(maxiter)):
        if nit % stepwise == 0:
            print('nit: ', nit)
            print('x: ', xx)
            print('tol: ', ff)
            print('+-----+-----+-----+-----+-----+-----+-----+-----+-----+-----+-----+-----+-----+-----+-----+-----+-----+-----+\n\n')

        vf[nsims-1] = ff
        if np.linalg.norm(g0) < tol_g:
            ccode = True
            break

        g1 = B.T @ g0
        g = B @ (g1/np.linalg.norm(g1))
        normg = np.linalg.norm(g)

        r = 1
        cal = 0
        deltax = 0

        while r > 0 and cal <= 500:
            cal += 1
            x = x + hs * g
            deltax = deltax + hs * normg
            f, g1, tt = calcfg(x)

            if f > ff:
                ff = f
                xx = x

            if np.mod(cal, nh) == 0:
                hs = hs * q2
            r = g @ g1

        if cal > 500:
            ccode = False
            break

        if cal == 1:
            hs = hs * q1

        ncals = ncals + cal
        if deltax < tol_x:
            ccode = True
            break

        dg = B.T @ (g1 - g0)
        xi = dg / np.linalg.norm(dg)
        B = B + w * np.outer((B @ xi), xi)
        g0 = g1

        vf = np.roll(vf, 1)
        vf[0] = abs(ff - vf[0])

        if abs(ff) > 1:
            deltaf = np.sum(vf)/abs(ff)
        else:
            deltaf = np.sum(vf)

        if deltaf < tol_f:
            ccode = True
            break
        ccode = False

    return ccode, xx, ff, tt


def _tol_tsopt(model, a, b, grad, weight, x0, tol, maxiter, stepwise):
    return __tolsolvty(model, grad, a, b, weight=weight, x0=x0, maxiter=maxiter, stepwise=stepwise,
                       tol_f=tol, tol_x=tol, tol_g=tol)[:-1]


def _tol_iopt(model, a, b, grad, x0, tol, maxiter, stepwise):
    def tol_globopt(func, x0, grad, tol, maxiter):
        def insert(zeroS, Y, v, vmag, bmm, nit_mon):
            if zeroS:
                newcol = (Y, v.a)
                bslindex = bisect_left(KeyWrapper(L, key=lambda c: c[1]), newcol[1])
                L.insert(bslindex, newcol)
            else:
                g = zeros(n)
                for k in range(m):
                    zeroG = False
                    for l in range(n):
                        g[l] = -grad[k](a[l], Y) * sgn(bmm[l])
                        if 0 in g[l]:
                            zeroG = True
                            break

                    if not zeroG:
                        nit_mon += 1
                        y1, y2 = Y.copy, Y.copy
                        y1[k], y2[k] = Y[k].a, Y[k].b
                        v1, v2 = func(y1).a, func(y2).a
                        if v1 < v2:
                            Y = y1
                        else:
                            Y = y2

                v = func(Y).a
                newcol = (Y, v)
                bslindex = bisect_left(KeyWrapper(L, key=lambda c: c[1]), newcol[1])
                L.insert(bslindex, newcol)

            return Y, v, nit_mon

        times_mon = 0
        Y = x0.copy
        y = func(Y).a
        L = [(Y, y)]
        n, m = len(a), len(grad)

        nit_mon = 0
        nit = 1
        while func(Y).wid >= tol and nit <= maxiter:
            if nit % stepwise == 0:
                print('nit: ', nit)
                print('x: ', Y)
                print('tol: ', -func(Y))
                print('+-----+-----+-----+-----+-----+-----+-----+-----+-----+-----+-----+-----+-----+-----+-----+-----+-----+-----+\n\n')

            gg = np.array([np.max((grad[k](a, Y)).mag) * Y[k].wid for k in range(m)])
            l = np.argmax(gg)
            Y1 = L[0][0].copy
            Y2 = L[0][0].copy
            Y1[l], Y2[l] = Interval(Y[l].a, Y[l].mid, sortQ=False), Interval(Y[l].mid, Y[l].b, sortQ=False)

            del L[0]

            bmm1, bmm2 = bm - model(a, Y1), bm - model(a, Y2)
            mag1, mag2 = abs(bmm1), abs(bmm2)
            v1, v2 = -min(br - mag1), -min(br - mag2)

            zeroS1 = True if (bmm1.a <= 0).all() and (0 <= bmm1.b).all() else False
            zeroS2 = True if (bmm2.a <= 0).all() and (0 <= bmm2.b).all() else False

            Y1, v1, nit_mon = insert(zeroS1, Y1, v1, mag1, bmm1, nit_mon)
            Y2, v2, nit_mon = insert(zeroS2, Y2, v2, mag2, bmm2, nit_mon)

            Y = L[0][0]
            nit += 1

        return L[0][0], func(L[0][0]), nit


    br = b.rad
    bm = b.mid

    minus_tol_interval = lambda x: -min(br - abs(bm - model(a, x)))
    xx, ff, nit = tol_globopt(minus_tol_interval, x0, grad, tol, maxiter)
    success = nit <= maxiter

    return success, xx, -ff


def Tol(model, a, b, x=None, maxQ=False, grad=None, weight=None, x0=None, tol=1e-12, maxiter=2000, stepwise=float('inf')):

    if maxQ:
        if isinstance(x0, INTERVAL_CLASSES):
            return _tol_iopt(model, a, b, grad, x0, tol, maxiter, stepwise)
        else:
            return _tol_tsopt(model, a, b, grad, weight, x0, tol, maxiter, stepwise)

    else:
        br = b.rad
        bm = b.mid

        tol_interval = lambda x: min( br - abs(bm - model(a, x)) )
        tol_exact    = lambda x: np.min( br - (bm - model(a, x)).mag )

        if isinstance(x, INTERVAL_CLASSES):
            return tol_interval(x)
        else:
            if x is None:
                raise TypeError('It is necessary to specify at which point to calculate the recognizing functional.')
            return tol_exact(x)
